Matrix

Struct Matrix 

pub struct Matrix<const N: usize, const M: usize> {
    pub rows: [[f64; M]; N],
}
A matrix with N rows and M columns, allocated on the stack.

Example

use hoomd_linear_algebra::matrix::Matrix;

let a = Matrix {
    rows: [[-1.0, 4.0, -6.0], [2.0, -3.0, 1.0]],
};

Fields

rows: [[f64; M]; N]

The elements of the matrix

Implementations

Source

impl<const N: usize, const M: usize> Matrix<N, M>

pub fn transpose(&self) -> Matrix<M, N>

Interchange the rows and columns of matrix.

\mathbf{A}^\top_{ji} = \mathbf{A}_{ij}
Examples
use hoomd_linear_algebra::matrix::Matrix;

let m = Matrix {
    rows: [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]],
};
let m_t = m.transpose();
assert_eq!(m_t.rows, [[1.0, 4.0], [2.0, 5.0], [3.0, 6.0]]);
pub fn map_rows<F>(self, f: F) -> Self
where F: FnMut([f64; M]) -> [f64; M],

Apply a function to an Matrix by rows.

Examples
use hoomd_linear_algebra::{Full, GeneralMatrix, matrix::Matrix22};

let m = Matrix22::full(3.0);
assert_eq!(
    m.map_rows(|v| [v[0] + 2.0, v[1]]),
    Matrix22 {
        rows: [[5.0, 3.0], [5.0, 3.0]]
    }
);
pub fn map_columns<F>(self, f: F) -> Self
where F: FnMut([f64; N]) -> [f64; N],

Apply a function to an Matrix by columns.

Examples
use hoomd_linear_algebra::{Full, GeneralMatrix, matrix::Matrix22};

let m = Matrix22::full(3.0);
assert_eq!(
    m.map_columns(|v| [v[0] + 2.0, v[1]]),
    Matrix22 {
        rows: [[5.0, 5.0], [3.0, 3.0]]
    }
);
pub fn map_elements<F>(self, f: F) -> Self
where F: Fn(f64) -> f64,

Apply a function to Matrix elements.

Examples
use hoomd_linear_algebra::{Full, GeneralMatrix, matrix::Matrix33};

let m = Matrix33::full(3.0);
assert_eq!(m.map_elements(|x| x + 2.0), m + Matrix33::full(2.0));
pub fn iter_elements(&self) -> impl Iterator<Item = f64>

Iterate over every element in the Matrix.

The iterator yields all matrix elements in row major order.

Examples
use hoomd_linear_algebra::{SquareMatrix, matrix::Matrix22};

let x = Matrix22 {
    rows: [[1.0, 2.0], [3.0, 4.0]],
};

let mut iterator = x.iter_elements();
assert_eq!(iterator.next(), Some(1.0));
assert_eq!(iterator.next(), Some(2.0));
assert_eq!(iterator.next(), Some(3.0));
assert_eq!(iterator.next(), Some(4.0));
assert_eq!(iterator.next(), None);
pub fn iter_elements_mut(&mut self) -> impl Iterator<Item = &mut f64>

Iterate over every element in the Matrix.

The iterator yields all matrix elements in row major order.

Examples
use hoomd_linear_algebra::{SquareMatrix, matrix::Matrix22};

let mut x = Matrix22 {
    rows: [[1.0, 2.0], [3.0, 4.0]],
};

let x_copy = x.clone();
let mut iterator = x.iter_elements_mut();
iterator.for_each(|x| *x *= 2.0);
assert_eq!(x, x_copy * 2.0);
pub fn iter_rows(&self) -> impl Iterator<Item = [f64; M]>

Iterate over matrix rows.

Examples
use hoomd_linear_algebra::{SquareMatrix, matrix::Matrix22};

let x = Matrix22 {
    rows: [[1.0, 2.0], [3.0, 4.0]],
};
let mut iterator = x.iter_rows();
assert_eq!(iterator.next(), Some([1.0, 2.0]));
assert_eq!(iterator.next(), Some([3.0, 4.0]));
assert_eq!(iterator.next(), None);
pub fn fold_elements<B, F>(self, init: B, f: F) -> B
where F: FnMut(B, f64) -> B,

Folds every element into an accumulator by applying an operation.

fold_elements takes two arguments: an initial value, and a closure with two arguments: an ‘accumulator’, and an element. The closure returns the value that the accumulator should have for the next iteration.

The initial value is the value the accumulator will have on the first call. After applying this closure to every element of the flattened iterator, fold_elements returns the accumulator.

Examples
use hoomd_linear_algebra::{Full, GeneralMatrix, matrix::Matrix22};

let m = Matrix22::full(3.0);
// Sum the elements of a matrix
assert_eq!(m.fold_elements(0.0, |acc, x| acc + x), 3.0 * 4.0);
pub fn fold_rows<B, F>(self, init: B, f: F) -> B
where F: FnMut(B, [f64; M]) -> B,

Folds every row into an accumulator by applying an operation.

fold_rows takes two arguments: an initial value, and a closure with two arguments: an ‘accumulator’, and an element. The closure returns the value that the accumulator should have for the next iteration.

The initial value is the value the accumulator will have on the first call. After applying this closure to every element of the flattened iterator, fold_rows returns the accumulator.

Examples
use hoomd_linear_algebra::{GeneralMatrix, matrix::Matrix22};

let m = Matrix22 {
    rows: [[1.0, 2.0], [3.0, 4.0]],
};
// Average the columns of a matrix
let n_rows = m.n_rows() as f64;
assert_eq!(
    m.fold_rows([0.0; 2], |acc, x| [acc[0] + x[0], acc[1] + x[1]])
        .map(|x| x / n_rows),
    [(1.0 + 3.0) / 2.0, (2.0 + 4.0) / 2.0]
);
pub const fn n_rows(&self) -> usize

Get the number of rows in the Matrix.

Examples
use hoomd_linear_algebra::matrix::Matrix;

let m = Matrix {
    rows: [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]],
};
assert_eq!(m.n_rows(), 2);
pub const fn n_columns(&self) -> usize

Get the number of columns in the Matrix.

Examples
use hoomd_linear_algebra::matrix::Matrix;

let m = Matrix {
    rows: [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]],
};
assert_eq!(m.n_columns(), 3);
Source

impl<const N: usize> Matrix<N, N>

pub fn with_diagonal(diagonal: [f64; N]) -> Self

Construct a square matrix with the given diagonal.

diagonal[i] is the matrix element $A_{ii}$. All off-diagonal elements are 0.

Example
use hoomd_linear_algebra::matrix::Matrix;

let a = Matrix::with_diagonal([2.0, -3.0]);

assert_eq!(a.rows, [[2.0, 0.0], [0.0, -3.0]]);
pub fn determinant(&self) -> f64

Compute the signed hypervolume of the hyperparallelepiped defined by a matrix.

This implementation uses the Laplace expansion, which is optimal for small matrices but will be extremely slow for large matrices due to its $O(N!)$ complexity.

Example
use hoomd_linear_algebra::{SquareMatrix, matrix::Matrix22};

let identity = Matrix22::identity();
assert_eq!(identity.determinant(), 1.0);

let scaled = identity * 2.0;
assert_eq!(scaled.determinant(), 2.0 * 2.0);
pub fn trace(&self) -> f64

Compute the sum of diagonal elements of a square matrix.

Examples
use hoomd_linear_algebra::{SquareMatrix, matrix::Matrix22};

let identity = Matrix22::identity();
assert_eq!(identity.trace(), 2.0);

let scaled = identity * 3.0;
assert_eq!(scaled.trace(), 3.0 + 3.0);
pub fn powi(&self, n: u32) -> Self

Compute a matrix to an integer power

\mathbf{A}^n = \prod_{i=1}^n \mathbf{A}
Examples
use hoomd_linear_algebra::{Full, GeneralMatrix, MatMul, matrix::Matrix22};

let matrix = Matrix22::full(2.0);

assert_eq!(matrix.powi(2), matrix.matmul(&matrix));

assert_eq!(matrix.powi(2).powi(2), matrix.powi(4));
pub fn diagonal(&self) -> DiagonalMatrix<N>

Extract the diagonal elements from a square matrix.

This method returns a DiagonalMatrix<N> containing the diagonal elements of the input matrix, where the element at position (i, i) is taken from the input matrix. All off-diagonal elements are ignored.

Examples
use hoomd_linear_algebra::matrix::Matrix33;
let a = Matrix33 {
    rows: [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]],
};
let b = a.diagonal();
assert_eq!(b.elements, [1.0, 5.0, 9.0]);
Source

impl Matrix<2, 2>

pub fn svd(&self) -> (Self, DiagonalMatrix<2>, Self)

Decompose a Matrix22 into a rotation, a scaling, and a second rotation.

\mathbf{A} = \mathbf{U} \boldsymbol{\Sigma} \mathbf{V}^\intercal

This implementation is based on the math in Blinn 1996, and ensures good (but not optimal) numerical stability. For certain pathological inputs, preconditioning the matrix could provide a benefit in numerical stability.

svd sets all singular values to be positive.

Examples
use hoomd_linear_algebra::{
    MatMul,
    matrix::{DiagonalMatrix, Matrix22},
};
let m = Matrix22 {
    rows: [[1.0, 2.0], [3.0, 4.0]],
};
let (u, s, vt) = m.svd();
let m_recon = u.matmul(&s.to_dense()).matmul(&vt);
for i in 0..2 {
    for j in 0..2 {
        assert!((m.rows[i][j] - m_recon.rows[i][j]).abs() < 1e-9);
    }
}
Source

impl Matrix<3, 3>

pub fn svd(&self) -> (Self, DiagonalMatrix<3>, Self)

Decompose a Matrix33 into a rotation, a scaling, and a second rotation.

\mathbf{A} = \mathbf{U} \boldsymbol{\Sigma} \mathbf{V}^\top

This implementation is based on the method described by McAdams 2011, which is a fast variant of the Jacobi iteration method.

The method ensures that U and V are pure rotation matrices (determinant = 1). As a result, the third singular value may be negative. For a conventional SVD with non-negative singular values, the sign can be absorbed into U.

Examples
use hoomd_linear_algebra::{
    MatMul, SquareMatrix,
    matrix::{DiagonalMatrix, Matrix33},
};
let m = Matrix33 {
    rows: [[1.0, 2.0, 3.0], [0.0, 1.0, 4.0], [5.0, 6.0, 0.0]],
};
let (u, s, vt) = m.svd();
let m_recon = u.matmul(&s.to_dense()).matmul(&vt);
for i in 0..3 {
    for j in 0..3 {
        assert!((m.rows[i][j] - m_recon.rows[i][j]).abs() < 1e-9);
    }
}